/**

lab3.c

Kyung-suk Lhee

**/

#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <cs_grid.h>
#include "block.h"

#define CONST 20
#define TAG 1
#define ALL MPI_COMM_WORLD
#define ANY MPI_ANY_SOURCE
#define loc(i, j) *(bp->data + (j) + (i) * bp->jmax)
#define COL CS_GRID_COLMAJOR
#define ROW CS_GRID_ROWMAJOR


void gcf(double val, int rgb[]) {

  if ( val <= 0.0 ) {           /* 0.0 = black */
    rgb[0] = rgb[1] = rgb[2] = 0;
  } else if ( val <= 1.0 ) {    /* 1.0 = blue */
    rgb[0] = rgb[1] = 0;
    rgb[2] = 255 * val;
  } else if ( val <= 2.0 ) {    /* 2.0 = red */
    rgb[0] = 255 * (val - 1.0);
    rgb[1] = 0;
    rgb[2] = 255 * (2.0 - val);
  } else if ( val <= 3.0 ) {    /* 3.0 = yellow */
    rgb[0] = 255;
    rgb[1] = 255 * (val - 2.0);
    rgb[2] = 0;
  } else if ( val <= 4.0 ) {    /* 4.0 = white */
    rgb[0] = rgb[1] = 255;
    rgb[2] = 255 * (val - 3.0);
  } else {      /* white */
    rgb[0] = rgb[1] = rgb[2] = 255;
  }
}

Block* BlockCreate(int x, int y) {

  Block *bp;

  bp = (Block*)malloc(sizeof(Block));
  bp->imax = x;
  bp->jmax = y;
  bp->data = (double*)malloc(sizeof(double) * x * y);
  bp->sendbuff = (double*)malloc(sizeof(double) * y);
  bp->lbuff = (double*)malloc(sizeof(double) * y);
  bp->rbuff = (double*)malloc(sizeof(double) * y);
  return bp;
}

void BlockInit(Block *bp, int tb, int te, int n, int id) {

  int iend, jend, i, j;
  double norm;

  bp->tb = tb;
  bp->te = te;
  bp->nblk = id;
  bp->nblks = n;
  bp->lnbr = (id + n - 1) % n;
  bp->rnbr = (id + 1) % n;
  iend = bp->imax - 1;
  jend = bp->jmax - 1;

  for (i=0; i<=iend; i++)
    for (j=0; j<=jend; j++)
      loc(i, j) = 0.0;
  norm = 0.0;

  for (i=0; i<=iend; i++) {
    loc(i, 0) = CONST;
    loc(i, jend) = CONST;
    norm += fabs(loc(i, 0) + loc(i, jend));
  }

  if (bp->nblk == 0)
    for (j=1; j<jend; j++) {
      loc(0, j) = CONST;
      norm += fabs(loc(0, j));
    }

  if (bp->nblk == (bp->nblks - 1))
    for (j=1; j<jend; j++) {
      loc(iend, j) = CONST;
      norm += fabs(loc(iend, j));
    }

  bp->norm = norm / bp->imax * bp->jmax;
}

void BlockSendL(Block *bp) {

  int j;

  for (j=0; j<bp->jmax; j++)
    bp->sendbuff[j] = loc(0, j);
  MPI_Send(bp->sendbuff, bp->jmax, MPI_DOUBLE, bp->lnbr, TAG, ALL);
}

void BlockSendR(Block *bp) {

  int j;

  for (j=0; j<bp->jmax; j++)
    bp->sendbuff[j] = loc((bp->imax)-1, j);
  MPI_Send(bp->sendbuff, bp->jmax, MPI_DOUBLE, bp->rnbr, TAG, ALL);
}

void BlockRecvR(Block *bp) {

  MPI_Status status;

  MPI_Recv(bp->rbuff, bp->jmax, MPI_DOUBLE, bp->rnbr, TAG, ALL, &status);
}

void BlockRecvL(Block *bp) {

  MPI_Status status;

  MPI_Recv(bp->lbuff, bp->jmax, MPI_DOUBLE, bp->lnbr, TAG, ALL, &status);
}

void BlockStep(Block *bp) {

  int iend, jend, i, j;
  double u1, u2, norm;

  iend = bp->imax - 1;
  jend = bp->jmax - 1;
  norm = 0.0;

  if (bp->nblk != 0)
    for (j=1; j<jend; j++) {
      u1 = loc(0, j);
      u2 = (loc(1, j) + bp->lbuff[j] + loc(0, j+1) +
            loc(0, j-1) + u1) / 5.0;
      loc(0, j) = u2;
      norm += fabs(u2 - u1);
    }
  for (j=1; j<jend; j++)
    for (i=1; i<iend; i++) {
      u1 = loc(i, j);
      u2 = (loc(i+1, j) + loc(i-1, j) + loc(i, j-1) +
            loc(i, j+1) + u1) / 5.0;
      loc(i, j) = u2;
      norm += fabs(u2 - u1);
    }
  if (bp->nblk != bp->nblks - 1)
    for (j=1; j<jend; j++) {
      u1 = loc(iend, j);
      u2 = (bp->rbuff[j] + loc(iend-1, j) + loc(iend, j+1) +
            loc(iend, j-1) + u1) / 5.0;
      loc(iend, j) = u2;
      norm += fabs(u2 - u1);
    }
  bp->norm = norm / (bp->imax * bp->jmax);
  bp->tb = bp->tb + 1;
}

void BlockNorm(Block *bp) {

  MPI_Reduce(&(bp->norm), &(bp->gnorm), 1, MPI_DOUBLE, MPI_SUM, 0, ALL);
  if (bp->nblk == 0)
    printf("global norm is:%g\n", bp->gnorm / bp->nblks);
}

void BlockDestroy(Block *bp) {

  free(bp->data);
  free(bp->sendbuff);
  free(bp->lbuff);
  free(bp->rbuff);
  free(bp);
}

void main(int argc, char **argv) {

  int id, size, imax, jmax, tb, te, i, j, color, blk, tmp;
  Block *bp;
  double *data;
  CS_Grid grid1, grid2;
  CS_Window win;
  MPI_Status status;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(ALL, &size);
  MPI_Comm_rank(ALL, &id);

  if (argc < 4) return;
  imax = atoi(argv[1]);
  jmax = atoi(argv[2]);
  tb = atoi(argv[3]);
  te = atoi(argv[4]);

  bp = BlockCreate(imax/size, jmax);
  BlockInit(bp, tb, te, size, id);
  BlockNorm(bp);
  blk = imax/size * jmax;
  if (id == 0) {
    printf("************************  ok 1 ********************\n");
    data = (double*)malloc(sizeof(double) * imax * jmax);
    printf("************************  ok 2 ********************\n");
    memcpy(data, bp->data, sizeof(double) * blk);
    printf("************************  ok 3 ********************\n");
    for (i=blk, j=1; j<size; i+=blk, j++) {
      MPI_Send(&tmp, 1, MPI_INT, j, TAG, ALL);
      printf("************************  send ********************\n");
      MPI_Recv(data+i, blk, MPI_DOUBLE, j, TAG, ALL, &status);
      printf("************************  recv ********************\n");
    }
    printf("************************  ok 4 ********************\n");
    grid1 = cs_grid_open(imax, jmax, gcf, "initial grid");
    printf("************************  ok 5 ********************\n");
    cs_grid_show(grid1, 0, 0, imax, jmax, data, COL);
    printf("************************  ok 6 ********************\n");
  }
  else {
    MPI_Recv(&tmp, 1, MPI_INT, ANY, TAG, ALL, &status);
    MPI_Send(bp->data, blk, MPI_DOUBLE, 0, TAG, ALL);
  }
  while (bp->tb <= bp->te) {
    BlockSendL(bp);
    BlockRecvR(bp);
    BlockSendR(bp);
    BlockRecvL(bp);
    BlockStep(bp);
    BlockNorm(bp);
  }
  if (id == 0) {
    printf("************************  ok 7 ********************\n");
    memcpy(data, bp->data, sizeof(double) * blk);
    printf("************************  ok 8 ********************\n");
    for (i=blk, j=1; j<size; i+=blk, j++) {
      MPI_Send(&tmp, 1, MPI_INT, j, TAG, ALL);
      MPI_Recv(data+i, blk, MPI_DOUBLE, j, TAG, ALL, &status);
    }
    grid2 = cs_grid_open(imax, jmax, gcf, "result grid");
    printf("************************  ok 9 ********************\n");
    cs_grid_show(grid2, 0, 0, imax, jmax, data, COL);
    printf("************************  ok 10 ********************\n");
    win = cs_make_window ("white", 200, 300, 300, 75);
    cs_set_color ("midnightblue");
    cs_text_size (30);
    cs_printf (win, 5, 25, "click to exit");
    while (cs_mouse_btn (win) == 0) ;
    cs_remove_window(win);
    cs_grid_close(grid1);
    cs_grid_close(grid2);
  }
  else {
    MPI_Recv(&tmp, 1, MPI_INT, ANY, TAG, ALL, &status);
    MPI_Send(bp->data, blk, MPI_DOUBLE, 0, TAG, ALL);
  }

  BlockDestroy(bp);
  MPI_Finalize();
}

